#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import pandas as pd

booleanMap = {
    "True": True,
    "False": False,
    "TRUE": True,
    "FALSE": False,
    "true": True,
    "false": False,
}


def convert_df(df: pd.DataFrame) -> pd.DataFrame:
    for col in df.columns:
        all_bools = all(str(val) in booleanMap for val in df[col])
        if all_bools:
            df[col] = df[col].map(lambda x: booleanMap[str(x)])
    return df



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.


from uuid import UUID

from freighter import Payload

from synnax.ontology.payload import ID


class Group(Payload):
    key: UUID
    name: str

    @property
    def ontology_id(self) -> ID:
        return ID(key=str(self.key), type="group")



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import math
import time

from synnax.telem import Rate, TimeSpan, TimeStamp

RESOLUTION = (100 * TimeSpan.MICROSECOND).seconds


def _precise_sleep(dur: float | int):
    estimate = RESOLUTION * 10  # Initial overestimate
    mean = RESOLUTION * 10
    m2 = 0
    count = 1
    end_time = time.perf_counter() + dur
    nanoseconds = dur * 1e9
    while nanoseconds > estimate * 1e9:
        start_time = time.perf_counter()
        time.sleep(RESOLUTION)
        # Elapsed time in nanoseconds
        elapsed = (time.perf_counter() - start_time) * 1e9
        delta = elapsed - mean
        mean += delta / count
        m2 += delta * (elapsed - mean)
        estimate = mean + 1 * math.sqrt(m2 / count)
        count += 1

    # Busy wait for the last bit to ensure we sleep for the correct duration
    while time.perf_counter() < end_time:
        pass


def sleep(dur: Rate | TimeSpan | float | int, precise: bool = False):
    """Sleeps for the given duration, with the option to use a high-precision sleep
    that is more accurate than Python's default time.sleep implementation.


    :param dur: The duration to sleep for. The value can be a float or int representing
    the number of seconds to sleep for, a Rate object representing the rate at
    which to sleep (i.e. 1 Hz = 1 second), or a TimeSpan object representing the
    duration to sleep for.
    :param precise: Whether to use a more precise sleep implementation. This will
    use more CPU than the default sleep implementation, but will provide significantly
    higher precision. It uses Welford's algorithm to estimate the ideal time to sleep
    for the given duration.
    """
    dur = TimeSpan.from_seconds(dur).seconds
    if precise:
        return _precise_sleep(dur)
    return time.sleep(dur)


class Timer:
    """Timer is a class that implements a simple timer. It is used to measure the time
    elapsed since the timer was started. The timer uses the high-resolution performance
    counter to measure time. The timer can be started, reset, and the elapsed time can
    be queried.
    """

    _start: TimeStamp

    def __init__(self):
        self.reset()

    def elapsed(self) -> TimeSpan:
        """Returns the time elapsed since the timer was started."""
        return TimeSpan(time.perf_counter_ns() - self._start)

    def start(self):
        """Starts the timer."""
        self.reset()

    def reset(self):
        """Resets the timer to zero."""
        self._start = TimeStamp(time.perf_counter_ns())


class Loop:
    """Loop is a class that implements a rate-limited loop. It is used to ensure that a
    block of code is executed at a fixed rate. The loop will sleep for the remainder of
    the interval if the block of code executes faster than the interval. If the block of
    code takes longer than the interval, the loop will skip sleeping to ensure that the
    block of code is executed as soon as possible. The loop keeps track of the average
    execution time and uses this to correct for any drift in the interval.
    """

    _timer: Timer
    interval: TimeSpan
    """The interval at which to run the loop."""
    counter: int = 0
    """The number of iterations that have been run."""
    average: TimeSpan = TimeSpan(0)
    """The average execution time of the loop."""
    _correction: TimeSpan = TimeSpan(0)

    def __init__(self, interval: Rate | TimeSpan | float | int, precise: bool = False):
        """Creates a new Loop object with the given interval and precision.

        :param interval: The interval at which to run the loop. This can be a float or
                int representing the number of seconds between each iteration, a Rate
                object representing the rate at which to run the loop (i.e. 1 Hz = 1
                second per iteration), or a TimeSpan object representing the duration
                between each iteration.
        :param precise: Whether to use a more precise sleep implementation. This will
                use more CPU than the default sleep implementation, but will provide
                significantly higher precision. It uses Welford's algorithm to estimate
                the ideal time to sleep for the given duration.
        """
        self._timer = Timer()
        self.interval = TimeSpan.from_seconds(interval)
        self.precise = precise
        self._last = time.perf_counter_ns()

    def __enter__(self):
        self._last = time.perf_counter_ns()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __iter__(self):
        return self

    def __next__(self):
        elapsed = self._timer.elapsed()
        if elapsed < self.interval:
            sleep_for = self.interval - elapsed - self._correction
            if sleep_for < TimeSpan(0):
                sleep_for = TimeSpan(0)
            sleep(sleep_for, self.precise)
        self.counter += 1
        self.average = TimeSpan(
            (self.average * (self.counter - 1) + self._timer.elapsed()) / self.counter
        )
        self._correction = self.average - self.interval
        self._timer.reset()

    def __call__(self):
        return self.__next__()

    def wait(self) -> True:
        """Waits for the next iteration of the loop, automatically sleeping for the
        remainder of the interval if the calling block of code executes faster than the
        interval.
        """
        self.__next__()
        return True



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

__version__ = "0.3.0"

from synnax.access import Policy, PolicyClient
from synnax.channel import Channel
from synnax.exceptions import (
    AuthError,
    ConfigurationError,
    ContiguityError,
    ControlError,
    ExpiredToken,
    InvalidToken,
    MultipleFoundError,
    NotFoundError,
    PathError,
    QueryError,
    RouteError,
    UnauthorizedError,
    UnexpectedError,
    ValidationError,
)
from synnax.framer import (
    AUTO_SPAN,
    Frame,
    Iterator,
    Streamer,
    Writer,
    WriterMode,
)
from synnax.hardware import Client as HardwareClient
from synnax.hardware.device import Device
from synnax.hardware.rack import Rack
from synnax.hardware.task import Task, TaskStatus, TaskStatusDetails
from synnax.options import SynnaxOptions
from synnax.ranger import Range
from synnax.synnax import Synnax
from synnax.telem import (
    Authority,
    Bounds,
    CrudeAuthority,
    CrudeDataType,
    CrudeDensity,
    CrudeRate,
    CrudeTimeSpan,
    CrudeTimeStamp,
    DataType,
    Density,
    Rate,
    Series,
    Size,
    TimeRange,
    TimeSpan,
    TimeSpanUnits,
    TimeStamp,
    convert_time_units,
    elapsed_seconds,
)
from synnax.timing import Loop, Timer, sleep
from synnax.user.payload import User

__all__ = [
    "AUTO_SPAN",
    "AuthError",
    "Authority",
    "Bounds",
    "Channel",
    "ConfigurationError",
    "ContiguityError",
    "ControlError",
    "convert_time_units",
    "CrudeAuthority",
    "CrudeDataType",
    "CrudeDensity",
    "CrudeRate",
    "CrudeTimeSpan",
    "CrudeTimeStamp",
    "DataType",
    "Density",
    "Device",
    "elapsed_seconds",
    "ExpiredToken",
    "PathError",
    "Frame",
    "HardwareClient",
    "Iterator",
    "InvalidToken",
    "Loop",
    "MultipleFoundError",
    "NotFoundError",
    "Policy",
    "PolicyClient",
    "QueryError",
    "Rack",
    "Range",
    "Rate",
    "RouteError",
    "Series",
    "Size",
    "sleep",
    "Streamer",
    "Synnax",
    "SynnaxOptions",
    "Task",
    "TaskStatus",
    "TaskStatusDetails",
    "Timer",
    "TimeRange",
    "TimeSpan",
    "TimeSpanUnits",
    "TimeStamp",
    "UnauthorizedError",
    "UnexpectedError",
    "User",
    "ValidationError",
    "Writer",
    "WriterMode",
]



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import overload

import pandas as pd
from alamos import NOOP, Instrumentation
from freighter import AsyncStreamClient, StreamClient, UnaryClient, WebsocketClient

from synnax.channel.payload import (
    ChannelKey,
    ChannelKeys,
    ChannelName,
    ChannelNames,
    ChannelParams,
    ChannelPayload,
    normalize_channel_params,
)
from synnax.channel.retrieve import ChannelRetriever
from synnax.exceptions import QueryError
from synnax.framer.adapter import ReadFrameAdapter, WriteFrameAdapter
from synnax.framer.deleter import Deleter
from synnax.framer.frame import CrudeFrame, Frame
from synnax.framer.iterator import Iterator
from synnax.framer.streamer import AsyncStreamer, Streamer
from synnax.framer.writer import CrudeWriterMode, Writer, WriterMode
from synnax.ontology import ID
from synnax.telem import (
    CrudeSeries,
    CrudeTimeStamp,
    MultiSeries,
    TimeRange,
    TimeSpan,
)
from synnax.telem.control import Authority, CrudeAuthority

ontology_type = ID(type="framer")


class Client:
    """FramerClient provides interfaces for reading and writing segmented
    telemetry from a Synnax Cluster. SegmentClient should not be instantiated
    directly, but rather used through the synnax.Synnax class.
    """

    __stream_client: WebsocketClient
    __async_client: AsyncStreamClient
    __unary_client: UnaryClient
    __channels: ChannelRetriever
    __deleter: Deleter
    instrumentation: Instrumentation

    def __init__(
        self,
        stream_client: WebsocketClient,
        async_client: AsyncStreamClient,
        unary_client: UnaryClient,
        retriever: ChannelRetriever,
        deleter: Deleter,
        instrumentation: Instrumentation = NOOP,
    ):
        self.__stream_client = stream_client
        self.__async_client = async_client
        self.__unary_client = unary_client
        self.__channels = retriever
        self.__deleter = deleter
        self.instrumentation = instrumentation

    def open_writer(
        self,
        start: CrudeTimeStamp,
        channels: ChannelParams,
        authorities: CrudeAuthority | list[CrudeAuthority] = Authority.ABSOLUTE,
        *,
        name: str = "",
        strict: bool = False,
        suppress_warnings: bool = True,
        mode: CrudeWriterMode = WriterMode.PERSIST_STREAM,
        err_on_unauthorized: bool = False,
        enable_auto_commit: bool = False,
        auto_index_persist_interval: TimeSpan = 1 * TimeSpan.SECOND,
        err_on_extra_chans: bool = True,
        use_experimental_codec: bool = True,
    ) -> Writer:
        """Opens a new writer on the given channels.

        :param start: Sets the starting timestamp for the first sample in the writer. If
        this timestamp overlaps with existing data for ANY of the provided channels,
        the writer will fail to open.
        :param channels: The channels to write to. This can be a single channel name,
        a list of channel names, a single channel key, or a list of channel keys.
        :param authorities: The control authority to set for each channel on the writer.
        Defaults to absolute authority. If not working with concurrent control,
        it's best to leave this as the default.
        :param name: The name of the writer used in control subject.
        :param strict: Sets whether the writer will fail to write if the data for a
        particular channel does not exactly match this data type. When False,
        the default, the writer will automatically convert the data to the correct
        type if possible.
        :param suppress_warnings: Suppress various print warnings that may be emitted
        by the writer.
        :param mode: sets the persistence and streaming mode of the writer. The default
        mode is WriterModePersistStream. See the WriterMode documentation for more.
        :param err_on_unauthorized: sets whether the writer should return an error if
        it attempts to write to a channel it does not have control over.
        :param enable_auto_commit: determines whether the writer will automatically
        commit. If enable_auto_commit is true, then the writer will commit after each
        write, and will flush that commit to index after the specified
        auto_index_persist_interval.
        :param auto_index_persist_interval: interval at which commits to the index will
        be persisted. To persist every commit to guarantee minimal loss of data, set
        auto_index_persist_interval to AlwaysAutoIndexPersist.
        """
        adapter = WriteFrameAdapter(
            retriever=self.__channels,
            err_on_extra_chans=err_on_extra_chans,
            strict_data_types=strict,
            suppress_warnings=suppress_warnings,
        )
        adapter.update(channels)
        return Writer(
            start=start,
            adapter=adapter,
            client=self.__stream_client,
            authorities=authorities,
            name=name,
            mode=mode,
            err_on_unauthorized=err_on_unauthorized,
            enable_auto_commit=enable_auto_commit,
            auto_index_persist_interval=auto_index_persist_interval,
            use_experimental_codec=use_experimental_codec,
        )

    def open_iterator(
        self,
        tr: TimeRange,
        channels: ChannelParams,
        chunk_size: int = 1e5,
    ) -> Iterator:
        """Opens a new iterator over the given channels within the provided time range.

        :param channels: A list of channel keys to iterator over.
        :param tr: A time range to iterate over.
        :param chunk_size: The number of samples to read in a chunk with AutoSpan. Defaults to 100000
        :returns: An Iterator over the given channels within the provided time
        range. See the Iterator documentation for more.
        """
        adapter = ReadFrameAdapter(self.__channels)
        adapter.update(channels)
        return Iterator(
            tr=tr,
            adapter=adapter,
            client=self.__stream_client,
            chunk_size=chunk_size,
            instrumentation=self.instrumentation,
        )

    @overload
    def write(
        self,
        start: CrudeTimeStamp,
        frame: CrudeFrame,
        strict: bool = False,
    ): ...

    @overload
    def write(
        self,
        start: CrudeTimeStamp,
        channel: ChannelKey | ChannelName | ChannelPayload,
        data: CrudeSeries,
        strict: bool = False,
    ):
        """Writes telemetry to the given channel starting at the given timestamp.

        :param channel: The key of the channel to write to.
        :param start: The starting timestamp of the first sample in data.
        :param data: The telemetry to write to the channel.
        :returns: None.
        """
        ...

    @overload
    def write(
        self,
        start: CrudeTimeStamp,
        channel: ChannelKeys | ChannelNames | list[ChannelPayload],
        series: list[CrudeSeries],
        strict: bool = False,
    ): ...

    def write(
        self,
        start: CrudeTimeStamp,
        channels: ChannelParams | ChannelPayload | list[ChannelPayload] | CrudeFrame,
        series: CrudeSeries | list[CrudeSeries] | None = None,
        strict: bool = False,
    ):
        parsed_channels = list()
        if isinstance(channels, (list, ChannelKey, ChannelPayload, ChannelName)):
            parsed_channels = channels
        elif isinstance(channels, dict):
            parsed_channels = list(channels.keys())
        elif isinstance(channels, Frame):
            parsed_channels = channels.channels
        elif isinstance(channels, pd.DataFrame):
            parsed_channels = list(channels.columns)
        with self.open_writer(
            start=start,
            channels=parsed_channels,
            strict=strict,
            mode=WriterMode.PERSIST,
            err_on_unauthorized=True,
            enable_auto_commit=True,
            auto_index_persist_interval=TimeSpan.MAX,
        ) as w:
            w.write(channels, series)

    @overload
    def read(
        self,
        tr: TimeRange,
        channels: ChannelKeys | ChannelNames,
    ) -> Frame: ...

    @overload
    def read(
        self,
        tr: TimeRange,
        channels: ChannelKey | ChannelName,
    ) -> MultiSeries: ...

    def read(
        self,
        tr: TimeRange,
        channels: ChannelParams,
    ) -> MultiSeries | Frame:
        """
        Reads telemetry from the channel between the two timestamps.

        :param tr: The time range to read from.
        :param channels: The key or name of the channel to read from.

        :returns: A tuple where the first item is a numpy array containing the telemetry
        and the second item is the time range occupied by that array.
        """
        normal = normalize_channel_params(channels)
        frame = self._read_frame(tr, channels)
        if len(normal.channels) > 1:
            return frame
        series = frame.get(normal.channels[0], None)
        if series is None:
            raise QueryError(
                f"""No data found for channel {normal.channels[0]} between {tr}"""
            )
        return series

    def open_streamer(
        self,
        channels: ChannelParams,
        downsample_factor: int = 1,
        use_experimental_codec: bool = True,
    ) -> Streamer:
        """Opens a new streamer on the given channels. The streamer will immediately
        being receiving frames of data from the given channels.

        :param channels: The channels to stream from. This can be a single channel name,
        a list of channel names, a single channel key, or a list of channel keys.

        :param downsample_factor: The downsample factor to use for the streamer.
        """
        adapter = ReadFrameAdapter(self.__channels)
        adapter.update(channels)
        return Streamer(
            adapter=adapter,
            client=self.__stream_client,
            downsample_factor=downsample_factor,
            use_experimental_codec=use_experimental_codec,
        )

    async def open_async_streamer(
        self, channels: ChannelParams, downsample_factor: int = 1
    ) -> AsyncStreamer:
        adapter = ReadFrameAdapter(self.__channels)
        adapter.update(channels)
        s = AsyncStreamer(
            adapter=adapter,
            client=self.__async_client,
            downsample_factor=downsample_factor,
        )
        await s._open()
        return s

    def delete(self, channels: ChannelParams, tr: TimeRange) -> None:
        """
        delete deletes data in the specified channels in the specified time range.
        Note that the time range is start-inclusive and end-exclusive.
        Also note that deleting all data in a channel does not delete the channel; to
        delete a channel, use client.channels.delete().
        :param channels: channels to delete data from.
        :param tr: time range to delete data from.
        """
        self.__deleter.delete(channels, tr)

    def _read_frame(
        self,
        tr: TimeRange,
        channels: ChannelParams,
    ) -> Frame:
        aggregate = Frame()
        with self.open_iterator(tr, channels) as it:
            for fr in it:
                aggregate.append(fr)
        return aggregate



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

import json
import warnings
from contextlib import contextmanager
from typing import Protocol, overload
from uuid import uuid4

from alamos import NOOP, Instrumentation
from freighter import Empty, Payload, UnaryClient, send_required
from pydantic import BaseModel, ValidationError

from synnax import UnexpectedError
from synnax.exceptions import ConfigurationError
from synnax.framer import Client as FrameClient
from synnax.hardware.rack import Client as RackClient
from synnax.hardware.rack import Rack
from synnax.hardware.task.payload import TaskPayload, TaskStatus
from synnax.status import ERROR_VARIANT, SUCCESS_VARIANT
from synnax.telem import TimeSpan, TimeStamp
from synnax.util.normalize import check_for_none, normalize, override


class _CreateRequest(Payload):
    tasks: list[TaskPayload]


_CreateResponse = _CreateRequest


class _DeleteRequest(Payload):
    keys: list[int]


class _RetrieveRequest(Payload):
    rack: int | None = None
    keys: list[int] | None = None
    names: list[str] | None = None
    types: list[str] | None = None
    include_status: bool = False


class _RetrieveResponse(Payload):
    tasks: list[TaskPayload] | None = None


_CREATE_ENDPOINT = "/hardware/task/create"
_DELETE_ENDPOINT = "/hardware/task/delete"
_RETRIEVE_ENDPOINT = "/hardware/task/retrieve"

_TASK_STATE_CHANNEL = "sy_task_status"
_TASK_CMD_CHANNEL = "sy_task_cmd"


class Task:
    key: int = 0
    name: str = ""
    type: str = ""
    config: str = ""
    snapshot: bool = False
    status: TaskStatus | None = None
    _frame_client: FrameClient | None = None

    def __init__(
        self,
        *,
        key: int = 0,
        rack: int = 0,
        name: str = "",
        type: str = "",
        config: str = "",
        snapshot: bool = False,
        status: TaskStatus | None = None,
        _frame_client: FrameClient | None = None,
    ):
        if key == 0:
            key = (rack << 32) + 0
        self.key = key
        self.name = name
        self.type = type
        self.config = config
        self.snapshot = snapshot
        self.status = status
        self._frame_client = _frame_client

    def to_payload(self) -> TaskPayload:
        return TaskPayload(
            key=self.key,
            name=self.name,
            type=self.type,
            config=self.config,
        )

    def set_internal(self, task: Task):
        self.key = task.key
        self.name = task.name
        self.type = task.type
        self.config = task.config
        self.snapshot = task.snapshot
        self._frame_client = task._frame_client

    def execute_command(self, type_: str, args: dict | None = None) -> str:
        """Executes a command on the task and returns the unique key assigned to the
        command.

        :param type_: The type of command to execute.
        :param args: The arguments to pass to the command.
        :return: The unique key assigned to the command.
        """
        w = self._frame_client.open_writer(TimeStamp.now(), _TASK_CMD_CHANNEL)
        key = str(uuid4())
        w.write(
            _TASK_CMD_CHANNEL,
            [{"task": self.key, "type": type_, "key": key, "args": args}],
        )
        w.close()
        return str(key)

    def execute_command_sync(
        self,
        type_: str,
        args: dict | None = None,
        timeout: float | TimeSpan = 5,
    ) -> TaskStatus:
        """Executes a command on the task and waits for the driver to acknowledge the
        command with a state.

        :param type_: The type of command to execute.
        :param args: The arguments to pass to the command.
        :param timeout: The maximum time to wait for the driver to acknowledge the
        command before a timeout occurs.
        """
        with self._frame_client.open_streamer([_TASK_STATE_CHANNEL]) as s:
            key = self.execute_command(type_, args)
            while True:
                frame = s.read(TimeSpan.from_seconds(timeout).seconds)
                if frame is None:
                    raise TimeoutError(
                        f"timed out waiting for driver to acknowledge {type_} command"
                    )
                elif _TASK_STATE_CHANNEL not in frame:
                    warnings.warn("task - unexpected missing state in frame")
                    continue
                try:
                    status = TaskStatus.model_validate(frame[_TASK_STATE_CHANNEL][0])
                    if status.key == key:
                        return status
                except ValidationError as e:
                    raise UnexpectedError(
                        f"""
                    Received invalid task state from driver.
                    """
                    ) from e


class MetaTask(Protocol):
    key: int

    def to_payload(self) -> TaskPayload: ...

    def set_internal(self, task: Task): ...


class StarterStopperMixin:
    _internal: Task

    def start(self, timeout: float | TimeSpan = 5):
        """Starts the task and blocks until the Synnax cluster has acknowledged the
        command or the specified timeout has elapsed.

        :raises TimeoutError: If the timeout is reached before the Synnax cluster
            acknowledges the command.
        :raises Exception: If the Synnax cluster fails to start the task correctly.
        """
        self._internal.execute_command_sync("start", timeout=timeout)

    def stop(self, timeout: float | TimeSpan = 5):
        """Stops the task and blocks until the Synnax cluster has acknowledged the
        command or the specified timeout has elapsed.

        :raises TimeoutError: If the timeout is reached before the Synnax cluster
            acknowledges the command.
        :raises Exception: If the Synnax cluster fails to stop the task correctly.
        """
        self._internal.execute_command_sync("stop", timeout=timeout)

    @contextmanager
    def run(self, timeout: float | TimeSpan = 5):
        """Context manager that starts the task before entering the block and stops the
        task after exiting the block. This is useful for ensuring that the task is
        properly stopped even if an exception occurs during execution.
        """
        self.start(timeout)
        try:
            yield
        finally:
            self.stop(timeout)


class JSONConfigMixin(MetaTask):
    _internal: Task
    config: BaseModel

    @property
    def name(self) -> str:
        return self._internal.name

    @property
    def key(self) -> int:
        """Implements MetaTask protocol"""
        return self._internal.key

    def to_payload(self) -> TaskPayload:
        """Implements MetaTask protocol"""
        pld = self._internal.to_payload()
        pld.config = json.dumps(self.config.dict())
        return pld

    def set_internal(self, task: Task):
        """Implements MetaTask protocol"""
        self._internal = task


class Client:
    _client: UnaryClient
    _frame_client: FrameClient
    _default_rack: Rack | None
    _racks: RackClient
    instrumentation: Instrumentation = NOOP

    def __init__(
        self,
        client: UnaryClient,
        frame_client: FrameClient,
        rack_client: RackClient,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self._client = client
        self._frame_client = frame_client
        self._racks = rack_client
        self._default_rack = None
        self.instrumentation = instrumentation

    @overload
    def create(
        self,
        *,
        key: int = 0,
        name: str = "",
        type: str = "",
        config: str = "",
        rack: int = 0,
    ): ...

    @overload
    def create(self, tasks: Task) -> Task: ...

    @overload
    def create(self, tasks: list[Task]) -> list[Task]: ...

    def create(
        self,
        tasks: Task | list[Task] | None = None,
        *,
        key: int = 0,
        name: str = "",
        type: str = "",
        config: str = "",
        rack: int = 0,
    ) -> Task | list[Task]:
        is_single = True
        if tasks is None:
            tasks = [TaskPayload(key=key, name=name, type=type, config=config)]
        elif isinstance(tasks, Task):
            tasks = [tasks.to_payload()]
        else:
            is_single = False
            tasks = [t.to_payload() for t in tasks]
        for pld in tasks:
            self.maybe_assign_def_rack(pld, rack)
        req = _CreateRequest(tasks=tasks)
        res = send_required(self._client, _CREATE_ENDPOINT, req, _CreateResponse)
        st = self.sugar(res.tasks)
        return st[0] if is_single else st

    def maybe_assign_def_rack(self, pld: TaskPayload, rack: int = 0) -> Rack:
        if self._default_rack is None:
            # Hardcoded as this value for now. Will be changed once we have multi-rack
            # systems
            self._default_rack = self._racks.retrieve_embedded_rack()
        if pld is not None and pld.key == 0:
            if rack == 0:
                rack = self._default_rack.key
            pld.key = (rack << 32) + 0
        return pld

    def configure(self, task: MetaTask, timeout: float = 5) -> MetaTask:
        with self._frame_client.open_streamer([_TASK_STATE_CHANNEL]) as streamer:
            pld = self.maybe_assign_def_rack(task.to_payload())
            req = _CreateRequest(tasks=[pld])
            res = send_required(self._client, _CREATE_ENDPOINT, req, _CreateResponse)
            task.set_internal(self.sugar(res.tasks)[0])
            while True:
                frame = streamer.read(timeout)
                if frame is None:
                    raise TimeoutError(
                        "task - timeout waiting for driver to "
                        "acknowledge configuration"
                    )
                elif (
                    _TASK_STATE_CHANNEL not in frame
                    or len(frame[_TASK_STATE_CHANNEL]) == 0
                ):
                    warnings.warn("task - unexpected missing state in frame")
                    continue
                status = TaskStatus.model_validate(frame[_TASK_STATE_CHANNEL][0])
                if status.details.task != task.key:
                    continue
                if status.variant == SUCCESS_VARIANT:
                    break
                if status.variant == ERROR_VARIANT:
                    raise ConfigurationError(status.message)
        return task

    def delete(self, keys: int | list[int]):
        req = _DeleteRequest(keys=normalize(keys))
        send_required(self._client, _DELETE_ENDPOINT, req, Empty)

    @overload
    def retrieve(
        self,
        *,
        key: int | None = None,
        name: str | None = None,
        type: str | None = None,
    ) -> Task: ...

    @overload
    def retrieve(
        self,
        names: list[str] | None = None,
        keys: list[int] | None = None,
        types: list[str] | None = None,
    ) -> list[Task]: ...

    def retrieve(
        self,
        key: int | None = None,
        name: str | None = None,
        type: str | None = None,
        names: list[str] | None = None,
        keys: list[int] | None = None,
        types: list[str] | None = None,
    ) -> list[Task] | Task:
        is_single = check_for_none(names, keys, types)
        res = send_required(
            self._client,
            _RETRIEVE_ENDPOINT,
            _RetrieveRequest(
                keys=override(key, keys),
                names=override(name, names),
                types=override(type, types),
            ),
            _RetrieveResponse,
        )
        sug = self.sugar(res.tasks)
        return sug[0] if is_single else sug

    def sugar(self, tasks: list[Payload]):
        return [Task(**t.model_dump(), _frame_client=self._frame_client) for t in tasks]



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from freighter import Payload

from synnax.status import Status


class TaskStatusDetails(Payload):
    """
    Details about the status of a task.
    """

    task: int = 0
    """The key of the task."""
    running: bool = False
    """Whether the task is running."""
    data: dict | None = None
    """Arbitrary data about the task."""


TaskStatus = Status[TaskStatusDetails]
"""The status of a task."""


class TaskPayload(Payload):
    """A primitive task payload."""

    key: int = 0
    name: str = ""
    type: str = ""
    config: str = ""
    snapshot: bool = False
    status: TaskStatus | None = None



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import Any, Generic, NotRequired, TypedDict, Unpack, overload

from synnax.cli.console.protocol import Print, Prompt, R
from synnax.exceptions import ValidationError


class AskKwargs(TypedDict, Generic[R]):
    arg: NotRequired[R | None]
    arg_name: NotRequired[str]


class NoneDefaultAskKwargs(AskKwargs[R]):
    default: NotRequired[R | None]


class DefaultAskKwargs(AskKwargs[R]):
    default: R


class SugaredConsole:
    enabled: bool
    prompt: Prompt
    print: Print

    def __init__(self, prompt: Prompt, print: Print, enabled: bool = True):
        self.prompt = prompt
        self.print = print
        self.enabled = enabled

    def info(self, message: str) -> None:
        if self.enabled:
            self.print.info(message)

    def success(self, message: str) -> None:
        if self.enabled:
            self.print.success(message)

    def warn(self, message: str) -> None:
        if self.enabled:
            self.print.warn(message)

    def error(self, message: str) -> None:
        if self.enabled:
            self.print.error(message)

    def table(self, columns: list[str], rows: list[dict]) -> None:
        if self.enabled:
            self.print.table(columns, rows)

    @overload
    def ask(
        self,
        question: str,
        type_: type[R] | None = None,
        choices: list[R] | None = None,
        password: bool = False,
        **kwargs: Unpack[AskKwargs[R]],
    ) -> R: ...

    @overload
    def ask(
        self,
        question: str,
        type_: type[R] | None = None,
        choices: list[str] | None = None,
        password: bool = False,
        **kwargs: Unpack[DefaultAskKwargs[R]],
    ) -> R: ...

    @overload
    def ask(
        self,
        question: str,
        type_: type[R] | None = None,
        choices: list[R] | None = None,
        password: bool = False,
        **kwargs: Unpack[NoneDefaultAskKwargs[R]],
    ) -> R | None:
        ...

        ...

    def ask(
        self,
        question: str,
        type_: type[R] | None = None,
        choices: list[R] | None = None,
        password: bool = False,
        **kwargs: Unpack[NoneDefaultAskKwargs[R]],
    ) -> R | None:
        v, default, should_return, has_default = self._validate(kwargs)
        if should_return:
            return v
        v = self.prompt.ask(
            question=question,
            type_=type_,
            choices=choices,
            default=default,
            password=password,
        )
        if v is not None or has_default:
            return v
        if self.print is not None:
            self.print.error("You must provide a value.")
        return self.ask(question, type_, choices, **kwargs)

    def _validate(
        self, kwargs: NoneDefaultAskKwargs[R]
    ) -> tuple[R | None, R | None, bool, bool]:
        has_default = "default" in kwargs

        default = kwargs.get("default", None)
        arg = kwargs.get("arg", None)
        if arg is not None:
            return arg, default, True, has_default

        arg_name = kwargs.get("arg_name", None)
        if not self.enabled and not has_default:
            raise ValueError(f"Missing required argument: {arg_name}")

        return default, default, not self.enabled, has_default

    @overload
    def select(
        self, rows: list[R], type_: type[R] = str, **kwargs: Unpack[DefaultAskKwargs[R]]
    ) -> tuple[R, int]: ...

    @overload
    def select(
        self,
        rows: list[dict[str, Any]],
        type_: type[R] = str,
        columns: list[str] | None = None,
        key: str | None = None,
        **kwargs: Unpack[DefaultAskKwargs[R]],
    ) -> tuple[R, int]: ...

    @overload
    def select(
        self,
        rows: list[R] | list[dict[str, Any]],
        type_: type[R],
        columns: list[str] | None = None,
        **kwargs: Unpack[AskKwargs[R]],
    ) -> tuple[R, int]: ...

    @overload
    def select(
        self,
        rows: list[R] | list[dict[str, Any]],
        type_: type[R],
        columns: list[str] | None = None,
        **kwargs: Unpack[NoneDefaultAskKwargs[R]],
    ) -> tuple[R | None, int | None]: ...

    def select(
        self,
        rows: list[R] | list[dict[str, Any]],
        type_: type[R] = str,
        columns: list[str] | None = None,
        key: str | None = None,
        **kwargs: Unpack[NoneDefaultAskKwargs[R]],
    ) -> tuple[R | None, int | None]:
        """Prompts the user to select a row from a table.

        :param ctx: The current flow Context.
        :param columns: The columns of the table.
        :param rows: The rows of the table.
        :param allow_none: Whether to allow the user to select nothing.
        :param default: The default option to select. If a default is provided,
        allow_none is ignored.
        :returns: The index of the selected row.py or None if nothing was selected.
        """
        v, default, should_return, has_default = self._validate(kwargs)
        if len(rows) == 0:
            raise ValidationError("No rows to select from.")
        if isinstance(rows[0], dict) and key is None:
            raise ValidationError("Missing key argument.")
        _key: str = key or "value"

        _rows = list()
        default_idx = 0
        no_cols = columns is None
        _columns = columns or list()

        for i, row in enumerate(rows):
            if isinstance(row, dict):
                is_default = row[_key] == default
                _rows.append({"choice": str(i), **row})
                if no_cols:
                    for k in row.keys():
                        if k not in _columns:
                            _columns.append(k)
            else:
                is_default = row == default
                key = "value"
                if len(_columns) > 0:
                    key = _columns[0]
                _rows.append({"choice": str(i), key: row})
            if is_default:
                default_idx = len(_rows) - 1

        if len(_columns) == 0:
            _columns = ["value"]

        if "choice" not in _columns:
            _columns = ["choice"] + _columns

        if should_return:
            return v, default_idx

        self.table(columns=_columns, rows=_rows)
        i = self.ask(
            "Select an option #",
            int,
            choices=[i for i in range(len(rows))],
            default=default_idx,
        )

        if i is not None:
            r = rows[i]
            assert r is not None
            return (r[_key], i) if isinstance(r, dict) else (r, i)
        if has_default:
            return default, default_idx
        if self.print is not None:
            self.print.error("You must make a selection.")
        return self.select(type_, rows, columns, key, **kwargs)  # type: ignore



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from alamos import Instrumentation, trace
from freighter import Empty, Payload, UnaryClient, send_required

from synnax.channel.payload import (
    ChannelKeys,
    ChannelNames,
    ChannelParams,
    ChannelPayload,
    normalize_channel_params,
)
from synnax.channel.retrieve import CacheChannelRetriever


class _CreateRequest(Payload):
    channels: list[ChannelPayload]


_Response = _CreateRequest


class _DeleteRequest(Payload):
    keys: ChannelKeys | None = None
    names: ChannelNames | None = None


_CREATE_ENDPOINT = "/channel/create"
_DELETE_ENDPOINT = "/channel/delete"
_RENAME_ENDPOINT = "/channel/rename"


class _RenameRequest(Payload):
    keys: ChannelKeys
    names: ChannelNames


class ChannelWriter:
    _client: UnaryClient
    _cache: CacheChannelRetriever | None
    instrumentation: Instrumentation

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation,
        cache: CacheChannelRetriever | None,
    ):
        self._client = client
        self.instrumentation = instrumentation
        self._cache = cache

    @trace("debug")
    def create(
        self,
        channels: list[ChannelPayload],
    ) -> list[ChannelPayload]:
        req = _CreateRequest(channels=channels)
        res = send_required(self._client, _CREATE_ENDPOINT, req, _Response)
        if self._cache is not None:
            self._cache.set(res.channels)
        return res.channels

    @trace("debug")
    def delete(self, channels: ChannelParams) -> None:
        normal = normalize_channel_params(channels)
        req = _DeleteRequest(**{normal.variant: normal.channels})
        send_required(self._client, _DELETE_ENDPOINT, req, Empty)
        if self._cache is not None:
            self._cache.delete(normal.channels)

    @trace("debug")
    def rename(self, keys: ChannelKeys, names: ChannelNames) -> None:
        req = _RenameRequest(keys=keys, names=names)
        send_required(self._client, _RENAME_ENDPOINT, req, Empty)
        if self._cache is not None:
            self._cache.rename(keys, names)



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import csv
from pathlib import Path
from typing import Iterator

import pandas as pd
from pandas.io.parsers import TextFileReader

from synnax.io.df_converter import convert_df
from synnax.io.matcher import new_extension_matcher
from synnax.io.protocol import ChannelMeta, FileWriter, ReaderType, RowFileReader

CSVMatcher = new_extension_matcher(["csv"])


class CSVReader(CSVMatcher):  # type: ignore
    """A RowReader implementation for CSV files."""

    channel_keys: list[str] | None
    chunk_size: int

    __reader: TextFileReader | None
    _path: Path
    _channels: list[ChannelMeta] | None
    _row_count: int | None
    _calculated_skip_rows: bool
    _skip_rows: int

    # Doing a protocol implementation check here because
    # it's hard for pyright to handle factories that return
    # protocol classes.
    def _(self) -> RowFileReader:
        return self

    def __init__(
        self,
        path: Path,
        keys: list[str] | None = None,
        chunk_size: int | None = None,
    ):
        self._path = path
        self.channel_keys = keys
        self.chunk_size = chunk_size or int(1e4)
        self._channels = None
        self._row_count = None
        self._skip_rows = 0
        self._calculated_skip_rows = False
        self.__reader = None

    def __detect_delimiter(self) -> str:
        with open(self._path, "r") as file:
            sample = file.read(1024)
            dialect = csv.Sniffer().sniff(sample)
            return dialect.delimiter

    def seek_first(self):
        self.close()
        self.__reader = pd.read_csv(
            self._path,
            chunksize=self.chunk_size,
            usecols=self.channel_keys,
            header=0,
            skiprows=self.__get_skip_rows(),
            delimiter=self.__detect_delimiter(),
        )

    def __get_skip_rows(self) -> int | tuple[int, int]:
        if self._calculated_skip_rows:
            return self._skip_rows

        r = pd.read_csv(
            self._path,
            chunksize=1,
            usecols=self.channel_keys,
            delimiter=self.__detect_delimiter(),
        )
        self._skip_rows = 0

        while not self._calculated_skip_rows:
            try:
                df = next(r)
            except StopIteration:
                # If we reach the end of the file, just do a best effort calculation
                self._skip_rows = 0
                break
            if isinstance(df.iloc[0, 0], str):
                self._skip_rows += 1
            else:
                self._calculated_skip_rows = True

        self._calculated_skip_rows = True

        r.close()
        if self._skip_rows > 0:
            return 1, self._skip_rows
        return self._skip_rows

    def channels(self) -> list[ChannelMeta]:
        if not self._channels:
            cols = pd.read_csv(
                self._path, nrows=0, delimiter=self.__detect_delimiter()
            ).columns
            self._channels = [
                ChannelMeta(name=name.strip(), meta_data=dict()) for name in cols
            ]
        return self._channels

    def set_chunk_size(self, chunk_size: int):
        self.chunk_size = chunk_size

    def read(self) -> pd.DataFrame:
        return convert_df(next(self._reader))

    def __iter__(self) -> Iterator[pd.DataFrame]:
        return CSVReaderIterator(self._reader)

    @classmethod
    def type(cls) -> ReaderType:
        return ReaderType.Row

    def path(self) -> Path:
        return self._path

    def nsamples(self) -> int:
        if not self._row_count:
            self._row_count = estimate_row_count(self._path)
        return self._row_count * len(self.channels())

    @property
    def _reader(self) -> TextFileReader:
        if self.__reader is None:
            self.seek_first()
        assert self.__reader is not None
        return self.__reader

    def close(self):
        if self.__reader is not None:
            self.__reader.close()
        self.__reader = None


def estimate_row_count(path: Path) -> int:
    """Estimates the number of rows in a CSV file."""
    with open(path, "r") as f:
        f.readline()
        row = f.readline()
        row_size = len(row.encode("utf-8"))

    file_size = path.stat().st_size
    return (file_size // row_size) - 1


class CSVWriter(CSVMatcher):  # type: ignore
    """A Writer implementation for CSV files."""

    _path: Path
    _header: bool

    def __init__(
        self,
        path: Path,
    ):
        self._path = path
        self._header = True
        if path.exists():
            path.unlink()

    # Doing a protocol implementation check here because
    # it's hard for pyright to handle factories that return
    # protocol classes.
    def _(self) -> FileWriter:
        return self

    def write(self, df: pd.DataFrame):
        df.to_csv(self._path, index=False, mode="a", header=self._header)
        self._header = False

    def path(self) -> Path:
        return self._path

    def close(self):
        pass


class CSVReaderIterator:
    base: Iterator[pd.DataFrame]

    def __init__(self, base: Iterator[pd.DataFrame]):
        self.base = base

    def __iter__(self):
        return self

    def __next__(self):
        return convert_df(next(self.base))



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import Generic, TextIO

from pydantic import BaseModel

from synnax.cli.console.protocol import (
    Console,
    Print,
    Prompt,
    R,
    assign_default_ask_type,
)
from synnax.cli.console.rich import RichConsole


class Entry(BaseModel, Generic[R]):
    message: str | None = None
    columns: list[str] | None = None
    rows: list[dict] | None = None
    choices: list[R] | None = None
    default: R | None = None
    response: R | None = None
    type_: type[R] | None = None
    password: bool | None = None


class Output(BaseModel):
    entries: list[Entry]

    def __init__(self, entries: list[Entry] | None = None):
        super().__init__(entries=entries or list())

    def append(self, entry: Entry):
        assert self.entries is not None
        self.entries.append(entry)

    def write(self, f: TextIO):
        f.write(self.json())


class MockPrint:
    """A mock implementation of the Print protocol for testing purposes."""

    output: Output
    verbose: Console | None

    def __init__(self, output: Output, verbose: bool = False):
        """
        :param output: The output list to append entries to.
        """
        self.output = output
        self.verbose = None if not verbose else RichConsole()

    def _(self) -> Print:
        return self

    def info(self, message: str):
        self.output.append(Entry(message=message))
        if self.verbose is not None:
            self.verbose.info(message)

    def error(self, message: str):
        self.output.append(Entry(message=message))
        if self.verbose is not None:
            self.verbose.error(message)

    def warn(self, message: str):
        self.output.append(Entry(message=message))
        if self.verbose is not None:
            self.verbose.warn(message)

    def success(self, message: str):
        self.output.append(Entry(message=message))
        if self.verbose is not None:
            self.verbose.success(message)

    def table(self, columns: list, rows: list):
        self.output.append(Entry(columns=columns, rows=rows))
        if self.verbose is not None:
            self.verbose.table(columns, rows)


class MockPrompt:
    """A mock implementation of the Prompt protocol for testing purposes."""

    output: Output
    responses: list

    def __init__(self, output: Output, responses: list):
        """
        :param output: The output list to append entries to.
        :param responses: A list of responses to return in order. These responses
        must be valid for the type of prompt being used.
        """
        self.output = output
        self.responses = responses

    def _(self) -> Prompt:
        return self

    def ask(
        self,
        question: str,
        type_: type[R] | None = None,
        choices: list[R] | None = None,
        default: R | None = None,
        password: bool = False,
    ) -> R | None:
        e = Entry(
            message=question,
            choices=choices,
            default=default,
            type_=assign_default_ask_type(type_, choices, default),
            password=password,
        )
        e.response = self.responses.pop(0) if len(self.responses) > 0 else default
        if type(e.response) != e.type_:
            raise TypeError(
                f"""
                Mock Prompt: Invalid response type
                Question: {question}
                Expected type: {type_}
                Actual response: {e.response}
                """
            )
        return e.response


class MockConsole(MockPrint, MockPrompt):
    """A mock implementation of the Console protocol for testing purposes."""

    def __init__(
        self,
        output: Output = Output(),
        responses: list | None = None,
        verbose: bool = False,
    ):
        """
        :param output: The output list to append entries to.
        :param responses: A list of responses to return in order. These responses
        must be valid for the type of prompt being used.
        """
        MockPrint.__init__(self, output, verbose)
        MockPrompt.__init__(self, output, responses or list())



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import os
from pathlib import Path

import click

from synnax.cli.connect import connect_from_options, prompt_client_options
from synnax.cli.console.rich import RichConsole
from synnax.cli.flow import Context
from synnax.cli.warning import warning
from synnax.config import ClusterConfig, ClustersConfig, ConfigFile


@click.command()
@click.pass_context
def login(ctx: click.Context) -> None:
    """Logs the user into a Synnax cluster and saves the parameters to the configuration
    file.
    """
    warning(ctx)
    ctx = Context(console=RichConsole())
    options = prompt_client_options(ctx)
    synnax = connect_from_options(ctx, options)
    if synnax is None:
        return
    cfg = ClustersConfig(ConfigFile(Path(os.path.expanduser("~/.synnax"))))
    cfg.set(ClusterConfig(options=options))
    ctx.console.info(SUCCESSFUL_LOGIN)


SUCCESSFUL_LOGIN = """Saved credentials. You can now use the Synnax Client
without having to log in. To connect the cluster in a Python shell, use the following:

from synnax import Synnax
client = Synnax()
"""



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import Protocol, TypeVar

R = TypeVar("R", str, int, float, bool, None)


class Prompt(Protocol):
    """A protocol class for an entity that can prompt the user for input."""

    def ask(
        self,
        question: str,
        type_: type[R] | None = None,
        choices: list[R] | None = None,
        default: R | None = None,
        password: bool = False,
    ) -> R | None:
        """Asks the user a question and returns their response.

        :param question: The question to ask the user.
        :param choices: A list of choices the user can select from. If provided,
        the user will be prompted to select one of the choices. A response not
        in the list will be rejected.
        :param default: A default value to use if the user does not provide a
        response. If provided, the user can press enter to select the default
        value.
        :return: The user's response.
        """
        ...


class Print(Protocol):
    """A protocol class for an entity that can print messages to the console."""

    def info(
        self,
        message: str,
    ) -> None:
        """Prints an informational message to the console.

        :param message: The message to print.
        """
        ...

    def warn(
        self,
        message: str,
    ) -> None:
        """Prints a warning message to the console.

        :param message: The message to print.
        """
        ...

    def error(
        self,
        message: str,
    ) -> None:
        """Prints an error message to the console.

        :param message: The message to print.
        """
        ...

    def success(
        self,
        message: str,
    ) -> None:
        """Prints a success message to the console.

        :param message: The message to print.
        """
        ...

    def table(
        self,
        columns: list[str],
        rows: list[dict],
    ):
        """Prints a table to the console.

        :param columns: A list of column names.
        :param rows: A list of dictionaries, where each dictionary represents a row.py. The
        keys of the dictionary should match the column names.
        """
        ...


class Console(Prompt, Print, Protocol):
    """A protocol class for an entity that can print messages to the console and prompt
    the user for input.
    """

    ...


def assign_default_ask_type(
    type_: type[R] | None, choices: list[R] | None, default: R | None
) -> type[R]:
    """Assigns a default type to the ask function.

    :param type_: The type to assign.
    :param default: The default value.
    :param choices: The list of choices.
    :return: The type.
    """
    if type_ is None:
        if choices is not None:
            type_ = type(choices[0])
        elif default is not None:
            type_ = type(default)  # type: ignore
        else:
            type_ = str  # type: ignore
    return type_  # type: ignore



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.


from typing import Any


def normalize[T](*args: T | tuple[T] | list[T] | None) -> list[T]:
    """Flatten a list of lists into a single list.

    Args:
        *args: A list of lists to flatten.

    Returns:
        A flattened list.
    """
    results: list[T] = list()
    if args[0] is None:
        return results
    for arg in args:
        if isinstance(arg, (list, tuple)):
            results.extend(arg)
        elif arg is not None:
            results.append(arg)
    return results


def check_for_none(*args: Any) -> bool:
    """Check if any of the arguments are None.

    Args:
        *args: A list of arguments to check for None.

    Returns:
        True if any of the arguments are None, False otherwise.
    """
    return all(arg is None for arg in args)


def override[T](*args: T | tuple[T] | list[T] | None) -> list[T] | None:
    for arg in args:
        if arg is not None:
            return normalize(arg)
    return None



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

import struct
from typing import Dict, List, Union

from freighter import JSONCodec
from freighter.codec import Codec as FreighterCodec

from synnax.channel.payload import ChannelKey, ChannelKeys
from synnax.exceptions import ValidationError
from synnax.framer.frame import Frame, FramePayload
from synnax.telem import DataType, Series, TimeRange

ZERO_ALIGNMENTS_FLAG_POS = 5
EQUAL_ALIGNMENTS_FLAG_POS = 4
EQUAL_LENGTHS_FLAG_POS = 3
EQUAL_TIME_RANGES_FLAG_POS = 2
TIME_RANGES_ZERO_FLAG_POS = 1
ALL_CHANNELS_PRESENT_FLAG_POS = 0

TIME_RANGE_SIZE = 16
ALIGNMENT_SIZE = 8
DATA_LENGTH_SIZE = 4
KEY_SIZE = 4
FLAGS_SIZE = 1
SEQ_NUM_SIZE = 4


class CodecFlags:
    def __init__(self) -> None:
        self.eq_len: bool = True
        self.eq_tr: bool = True
        self.time_ranges_zero: bool = True
        self.all_channels_present: bool = True
        self.eq_align: bool = True
        self.zero_alignments: bool = True

    def encode(self) -> int:
        b = 0
        if self.eq_len:
            b |= 1 << EQUAL_LENGTHS_FLAG_POS
        if self.eq_tr:
            b |= 1 << EQUAL_TIME_RANGES_FLAG_POS
        if self.time_ranges_zero:
            b |= 1 << TIME_RANGES_ZERO_FLAG_POS
        if self.all_channels_present:
            b |= 1 << ALL_CHANNELS_PRESENT_FLAG_POS
        if self.eq_align:
            b |= 1 << EQUAL_ALIGNMENTS_FLAG_POS
        if self.zero_alignments:
            b |= 1 << ZERO_ALIGNMENTS_FLAG_POS
        return b

    @classmethod
    def decode(cls, b: int) -> CodecFlags:
        flags = cls()
        flags.eq_len = bool((b >> EQUAL_LENGTHS_FLAG_POS) & 1)
        flags.eq_tr = bool((b >> EQUAL_TIME_RANGES_FLAG_POS) & 1)
        flags.time_ranges_zero = bool((b >> TIME_RANGES_ZERO_FLAG_POS) & 1)
        flags.all_channels_present = bool((b >> ALL_CHANNELS_PRESENT_FLAG_POS) & 1)
        flags.eq_align = bool((b >> EQUAL_ALIGNMENTS_FLAG_POS) & 1)
        flags.zero_alignments = bool((b >> ZERO_ALIGNMENTS_FLAG_POS) & 1)
        return flags


class CodecState:
    keys: ChannelKeys
    data_types: Dict[ChannelKey, DataType]
    has_variable_data_types: bool

    def __init__(self, keys: ChannelKeys, data_types: List[DataType]) -> None:
        self.keys = sorted(keys)
        self.data_types = {k: dt for k, dt in zip(keys, data_types)}
        self.has_variable_data_types = any(dt.is_variable for dt in data_types)


class Codec:
    _has_variable_data_types: bool
    _seq_num: int
    _states: dict[int, CodecState]
    _curr_state: CodecState = None

    def __init__(
        self, keys: ChannelKeys = None, data_types: List[DataType] = None
    ) -> None:
        self._seq_num = 0
        self._states = dict()
        if keys is not None:
            self.update(keys, data_types)

    def update(self, keys: ChannelKeys, data_types: list[DataType]):
        self._seq_num += 1
        self._curr_state = CodecState(keys, data_types)
        self._states[self._seq_num] = self._curr_state

    def throw_if_not_updated(self, op_name: str):
        if self._curr_state is None:
            raise ValueError(
                "Codec has not been updated with keys and data types. "
                f"Please call update() before calling {op_name}()."
            )

    def encode(self, frame: Union[Frame, FramePayload], start_offset: int = 0) -> bytes:
        self.throw_if_not_updated("encode")
        pld = frame if isinstance(frame, FramePayload) else frame.to_payload()
        indices = sorted(range(len(pld.keys)), key=lambda i: pld.keys[i])
        sorted_keys = [pld.keys[i] for i in indices]
        sorted_series = [pld.series[i] for i in indices]

        flg = CodecFlags()
        flg.eq_len = not self._curr_state.has_variable_data_types
        curr_data_size = -1
        ref_tr = None
        ref_align = None
        byte_array_size = start_offset + FLAGS_SIZE + SEQ_NUM_SIZE

        if len(sorted_keys) != len(self._curr_state.keys):
            flg.all_channels_present = False
            byte_array_size += len(sorted_keys) * KEY_SIZE

        for i, ser in enumerate(sorted_series):
            key = sorted_keys[i]
            dt = self._curr_state.data_types.get(key, None)
            if dt is None:
                raise ValidationError(
                    f"encoder was provided a key {key} that is not in the codec state."
                )
            elif dt != ser.data_type:
                raise ValidationError(
                    f"data type {ser.data_type} for key {key} does not match codec state data type {dt}."
                )
            if curr_data_size == -1:
                curr_data_size = len(ser)
                ref_tr = ser.time_range
                ref_align = ser.alignment
            else:
                if len(ser) != curr_data_size:
                    flg.eq_len = False
                if ser.time_range != ref_tr:
                    flg.eq_tr = False
                if ser.alignment != ref_align:
                    flg.eq_align = False

            byte_array_size += len(ser.data)

        flg.time_ranges_zero = ref_tr is None or ref_tr == 0
        flg.zero_alignments = flg.eq_align and ref_align == 0
        byte_array_size += (int(flg.eq_len) or len(sorted_keys)) * DATA_LENGTH_SIZE
        if not flg.time_ranges_zero:
            byte_array_size += (int(flg.eq_tr) or len(sorted_keys)) * TIME_RANGE_SIZE
        if not flg.zero_alignments:
            byte_array_size += (int(flg.eq_align) or len(sorted_keys)) * ALIGNMENT_SIZE

        buffer = bytearray(byte_array_size)
        offset = start_offset

        buffer[offset] = flg.encode()
        offset += FLAGS_SIZE

        struct.pack_into("<I", buffer, offset, self._seq_num)
        offset += SEQ_NUM_SIZE

        if flg.eq_len:
            struct.pack_into("<I", buffer, offset, curr_data_size)
            offset += DATA_LENGTH_SIZE

        if flg.eq_tr and not flg.time_ranges_zero and ref_tr:
            struct.pack_into("<QQ", buffer, offset, ref_tr.start, ref_tr.end)
            offset += TIME_RANGE_SIZE

        if flg.eq_align and not flg.zero_alignments and ref_align:
            struct.pack_into("<Q", buffer, offset, ref_align)
            offset += ALIGNMENT_SIZE

        for i, ser in enumerate(sorted_series):
            k = sorted_keys[i]
            dt = self._curr_state.data_types.get(k, None)
            if dt is None:
                raise ValidationError(
                    f"Codec state does not contain data type for key {k}."
                )
            elif dt != ser.data_type:
                raise ValidationError(
                    f"Series data type {ser.data_type} for key {k} does not match codec state data type {dt}."
                )

            if not flg.all_channels_present:
                struct.pack_into("<I", buffer, offset, sorted_keys[i])
                offset += KEY_SIZE

            if not flg.eq_len:
                len_or_size = len(ser)
                if ser.data_type.is_variable:
                    len_or_size = len(ser.data)
                struct.pack_into("<I", buffer, offset, len_or_size)
                offset += DATA_LENGTH_SIZE

            buffer[offset : offset + len(ser.data)] = ser.data
            offset += len(ser.data)

            if not flg.eq_tr and not flg.time_ranges_zero:
                if ser.time_range is None:
                    tr = TimeRange.ZERO
                else:
                    tr = ser.time_range
                struct.pack_into("<QQ", buffer, offset, tr.start, tr.end)
                offset += TIME_RANGE_SIZE

            if not flg.eq_align and not flg.zero_alignments:
                struct.pack_into("<Q", buffer, offset, ser.alignment)
                offset += ALIGNMENT_SIZE

        return bytes(buffer)

    def decode(self, data: bytes, offset: int = 0) -> FramePayload:
        self.throw_if_not_updated("decode")
        buffer = memoryview(data)
        idx = offset
        flags = CodecFlags.decode(buffer[idx])
        idx += 1

        curr_seq_num = struct.unpack_from("<I", buffer, idx)[0]
        idx += SEQ_NUM_SIZE

        state = self._states.get(curr_seq_num)
        if state is None:
            return FramePayload()

        to_del = None
        for seq_num in self._states.keys():
            if seq_num < curr_seq_num:
                if to_del is None:
                    to_del = set()
                to_del.add(seq_num)

        if to_del is not None:
            for seq_num in to_del:
                del self._states[seq_num]

        data_len = 0
        start_time = 0
        end_time = 0
        alignment = 0

        if flags.eq_len:
            data_len = struct.unpack_from("<I", buffer, idx)[0]
            idx += DATA_LENGTH_SIZE

        if flags.eq_tr and not flags.time_ranges_zero:
            start_time, end_time = struct.unpack_from("<QQ", buffer, idx)
            idx += TIME_RANGE_SIZE

        if flags.eq_align and not flags.zero_alignments:
            alignment = struct.unpack_from("<Q", buffer, idx)[0]
            idx += ALIGNMENT_SIZE

        keys = list()
        series_list = list()

        for key in state.keys:
            if not flags.all_channels_present:
                if idx >= len(buffer):
                    break
                frame_key = struct.unpack_from("<I", buffer, idx)[0]
                if frame_key != key:
                    continue
                idx += KEY_SIZE
            data_type = state.data_types[key]
            curr_len = data_len
            if not flags.eq_len:
                curr_len = struct.unpack_from("<I", buffer, idx)[0]
                idx += DATA_LENGTH_SIZE

            data_byte_len = curr_len
            if not data_type.is_variable:
                data_byte_len = curr_len * data_type.density

            series_data = bytes(buffer[idx : idx + data_byte_len])
            idx += data_byte_len

            if flags.time_ranges_zero:
                tr = TimeRange.ZERO
            elif flags.eq_tr:
                tr = TimeRange(start=start_time, end=end_time)
            else:
                s, e = struct.unpack_from("<QQ", buffer, idx)
                tr = TimeRange(start=s, end=e)
                idx += TIME_RANGE_SIZE

            curr_alignment = alignment
            if not flags.eq_align and not flags.zero_alignments:
                curr_alignment = struct.unpack_from("<Q", buffer, idx)[0]
                idx += ALIGNMENT_SIZE

            keys.append(key)
            series_list.append(
                Series(
                    data_type=data_type,
                    data=series_data,
                    time_range=tr,
                    alignment=curr_alignment,
                )
            )

        return FramePayload(keys=keys, series=series_list)


LOW_PERF_SPECIAL_CHAR = 254
HIGH_PERF_SPECIAL_CHAR = 255
CONTENT_TYPE = "application/sy-framer"


class WSFramerCodec(FreighterCodec):
    def __init__(self, codec: Codec) -> None:
        self.codec = codec
        self.lower_perf_codec = JSONCodec()

    def content_type(self) -> str:
        return CONTENT_TYPE



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from synnax.status.status import (
    ERROR_VARIANT,
    INFO_VARIANT,
    SUCCESS_VARIANT,
    WARNING_VARIANT,
    Status,
    Variant,
)

__all__ = [
    "Variant",
    "Status",
    "SUCCESS_VARIANT",
    "INFO_VARIANT",
    "WARNING_VARIANT",
    "ERROR_VARIANT",
]



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import Generic, Literal, TypeVar
from uuid import uuid4

from freighter import Payload
from pydantic import Field

from synnax.telem import TimeStamp

SUCCESS_VARIANT = "success"
INFO_VARIANT = "info"
WARNING_VARIANT = "warning"
ERROR_VARIANT = "error"
DISABLED_VARIANT = "disabled"
LOADING_VARIANT = "loading"

Variant = Literal[
    "success",
    "info",
    "warning",
    "error",
    "disabled",
    "loading",
]
"""Represents the variant of a status message."""

D = TypeVar("D", bound=Payload)


class Status(Payload, Generic[D]):
    """A standardized payload used across Synnax."""

    key: str = Field(default_factory=lambda: str(uuid4()))
    """A unique key for the status."""
    variant: Variant
    """The variant of the status."""
    message: str
    """The message of the status."""
    description: str = ""
    """The description of the status."""
    time: TimeStamp = Field(default_factory=TimeStamp.now)
    """The time the status was created."""
    details: D
    """The details are customizable details for component specific statuses."""
